#include "../state.frag"
#include "../other/customExtras.sh"

USE_TEXTURE2D(normalMap); //linear name "Normal map"

USE_TEXTURE2D(diffuseMap); //srgb name "Base color map"
uniform int uDiffuseGrey; //bool name "Base color greyscale" default 0
uniform int uDiffuseAlpha; //bool name "Opacity" default 0
uniform float uDiffuseAlphaTest; //name "Alpha test value" default 0.5 min 0.0 max 1.0

//Spec map should be linear colorspace (acts as both metal and spec map, workflow is switched in the shader)
USE_TEXTURE2D(specularMap); //linear name "Specular map"
uniform int uSpecGrey; //bool name "Specular greyscale" default 1
uniform int uSpecWorkflow; //name "PBR workflow" default 1 labels "Metalness" "Specular"

USE_TEXTURE2D(roughnessMap); //linear name "Roughness map"

USE_TEXTURE2D(tintMaskMap); //linear name "Tint map"
uniform int uUseTint; //bool name "Use tint" default 1
uniform vec3 uTint0; //color name "Tint 0" default 0.13,0.03,0.005
uniform vec3 uTint1; //color name "Tint 1" default 0.032,0.032,0.032
uniform vec3 uTint2; //color name "Tint 2" default 0.0,0.12,0.19
uniform vec3 uTint3; //color name "Tint 3" default 0.35,0.35,0.35

USE_TEXTURE2D(emissiveMap); //srgb name "Emissive map"
uniform int uUseEmissive; //bool name "Use emissive" default 0
uniform vec3 uEmissiveLow; //color name "Emissive low" default 0.03,0.26,0.73
uniform vec3 uEmissiveHigh; //color name "Emissive high" default 0.21,0.71,1.0
uniform float uEmissiveAtten; //name "Emissive atten" default 3.0 min 0.0 max 100.0
uniform int uEmMask; //bool name "Animated emissive" default 0
uniform vec2 uEmissivePan; //name "Emissive pan" default 0.0,0.0 min -20.0,-20.0 max 20.0, 20.0
uniform int uRT; //bool name "Use real time" default 0
uniform int uEmFresnel; //bool name "Emissive fresnel" default 0
uniform float uEmFresnelPow; //name "Fresnel power" default 1.0 min 0.01 max 10.0

uniform int uValidate; //name "Validate PBR filter" default 0 min 0 max 2

//Helper functions
/*
float remapSpecToMetal(float F0)
{
	return saturate(F0 - 0.04) * (1.0/0.96);
}

float remapSpecToEta(float F0)
{
	float F0r = sqrt(saturate(F0));
	return min((1.0-F0r) * rcp(1.0 + F0r), 1.0/1.01);
}

float maxcomponent(vec3 v)
{
	return max( max( v.x, v.y ), v.z );
}
*/

vec3 Vec3ToGrey(vec3 i)
{
    float grey = 0.21 * i.r + 0.72 * i.g + 0.07 * i.b;
    if (i.g == 0 && i.b == 0) //in case it's a grayscale map (only has r)
    {
        i.rgb = i.rrr;
    }
    else i.rgb = vec3(grey, grey, grey);
    return i;
}

float Linear2sRGB(float x)
{
    return x <= 0.031308 ? 12.92 * x : 1.055 * pow(x, 0.41666) - 0.055;
}

float sRGB2linear(float x)
{
    return x <= 0.04045 ? x * 0.0773993808 : pow((x + 0.055) / 1.055, 2.4);
}

vec3 Linear2sRGBv3(vec3 rgb)
{
    return vec3(
        Linear2sRGB(rgb.r),
        Linear2sRGB(rgb.g),
        Linear2sRGB(rgb.b));
}

vec3 sRGB2linearV3(vec3 rgb)
{
    return vec3(
        sRGB2linear(rgb.r),
        sRGB2linear(rgb.g),
        sRGB2linear(rgb.b));
}


//Filter for compliance with PBR and tintmap needs
vec4 ValidateWFPBR (vec4 a, vec3 r, int isSpec)
{

	//Lowest allowed metal albedo
	const float aML = 0.52252155;

	//Per testing in both mset and EE; metal albedo values higher than this result in vantablack
	const float aMH = 0.65593;

	//Lowest allowed metalness value
	//const float mL = 0.83076987;

	//Initialize heatmap
	const vec4 vR = vec4(1.0,0.0,0.0,1.0);
	const vec4 vG = vec4(0.0,1.0,0.0,1.0);
	const vec4 vB = vec4(0.0,0.0,1.0,1.0);
	vec4 heatmap = vec4(1.0,1.0,1.0,1.0);
	
	//Values for Warframe maps, we want 0.5 sRGB diffuse with minor variation for consistent tinting
	const float aWFL = 0.119;
	const float aWFH = 0.35;
	const float sR = 0.08;
	const float specWFL = 0.447;


    float grey = 0.5;
    float s = 0.5;
    if (!uDiffuseGrey)
    {
	    grey = 0.21 * a.r + 0.71 * a.g + 0.07 * a.b;
    }
    else grey = a.r;
	
	//temp masks
	float fMaskL = 1.0;
	float fMaskH = 1.0;

    if (!uSpecGrey)
    {
	    s = 0.21 * r.r + 0.71 * r.g + 0.07 * r.b;
    }
    else s = r.r;

	float m = s;
	s = sRGB2linear(s);
    //s = Linear2sRGB(s);

	vec4 albedoHeat = vec4(1.0,1.0,1.0,1.0);
	vec4 diffuseHeat = vec4(1.0,1.0,1.0,1.0);
	vec4 metalHeat = vec4(1.0,1.0,1.0,1.0);
	vec4 specHeat = vec4(1.0,1.0,1.0,1.0);
	
	//diffuse heatmap
	fMaskL = mix(0.0, 1.0, smoothstep(aWFL - sR, aWFL + sR, grey));
	diffuseHeat = mix(vR, vG, fMaskL);
	fMaskH = mix(0.0, 1.0, smoothstep(aWFH - sR, aWFH + sR, grey));
	diffuseHeat = mix(diffuseHeat, vB, fMaskH);

	//specular heatmap
	fMaskL = mix(0.0, 1.0, smoothstep(aML - sR, aML + sR, s));
	specHeat = mix(vR, vG, fMaskL);
	fMaskH = mix(0.0, 1.0, step(aMH, s));
	specHeat = mix(specHeat, vB, fMaskH);

	//albedo heatmap
	fMaskL = mix(0.0, 1.0, smoothstep(aWFL - sR, aWFL + sR, grey));
	albedoHeat = mix(vR, vG, fMaskL);
	fMaskH = mix(0.0, 1.0, smoothstep(aWFH - sR, aMH + sR, grey));
	albedoHeat = mix(albedoHeat, vB, fMaskH);

	//metal heatmap
	fMaskL = mix(0.0, 1.0, smoothstep(aML, aML + sR, m));
	metalHeat = mix(vR, vG, fMaskL);
	fMaskH = mix(0.0, 1.0, step(1.0, m));
	metalHeat = mix(metalHeat, vB, fMaskH);

	//combine diffuse-spec heat
	if (isSpec) heatmap = mix(specHeat, diffuseHeat, smoothstep(0.01, 0.1, grey));
	//combine albedo-metal heat
	else heatmap = mix(albedoHeat, metalHeat, m);

	return heatmap;
}

float fTime(int useRT)
{
	float t = uCustomAnimationTime;
	if (useRT) t = uCustomTime;
	return t;
}

//main PBR function
void WarframePBR ( inout FragmentState s )
{
    vec3 T = s.vertexTangent;
	vec3 B = s.vertexBitangent;
	vec3 N = s.vertexNormal;

	float yCoeff = -1.0; //DirectX format
	vec3 nrmMap = textureMaterial(normalMap, s.vertexTexCoord).rgb;
	vec3 nrm = nrmMap.xyz * 2.0 - vec3(1.0, 1.0, 1.0);

	nrm.y *= yCoeff;
	nrm = nrm.x * T + nrm.y * B + nrm.z * N;
		
	s.normal = normalize(nrm);

    //load base maps, mindful of greyscale
    s.albedo = textureMaterial(diffuseMap, s.vertexTexCoord);
    //ensure albedo.rgb == albedo.rrr (greyscale) if desired by the user
    s.albedo.rgb = uDiffuseGrey ? Vec3ToGrey(s.albedo.rgb) : s.albedo.rgb;

    vec3 specColor = texture2D(specularMap, s.vertexTexCoord).rgb;
    specColor = uSpecGrey ? Vec3ToGrey(specColor) : specColor;
    float m = specColor.r;
    
    s.fresnel = 0.0;

    float g = 1.0 - textureMaterial(roughnessMap, s.vertexTexCoord).r;

    //PBR validation filter check
    if (uValidate > 0)
    {
        vec4 vHeat = ValidateWFPBR(s.albedo, specColor, uSpecWorkflow);
        s.albedo = vHeat;
        if(uValidate == 2)
        {
            s.emissiveLight = vHeat.rgb;
            s.gloss = 0.0;
        }
        //Return early so we're left with heatmap as the output
        return;
    }

    //calculate tint
    vec4 resultTint = vec4(1.0,1.0,1.0,1.0);
    if (uUseTint)   
    {
        vec4 tintMask = texture2D(tintMaskMap, s.vertexTexCoord).rgba;
		resultTint = mix(vec4(0.5,0.5,0.5,0.5), vec4(uTint0.rgb,1), tintMask.r);
		resultTint = mix(resultTint, vec4(uTint1.rgb,1), tintMask.g);
		resultTint = mix(resultTint, vec4(uTint2.rgb,1), tintMask.b);
		resultTint = mix(resultTint, vec4(uTint3.rgb,1), tintMask.a);
		resultTint *= 2.0;
    }

    s.albedo.rgb = s.albedo.rgb * resultTint.rgb;

    if(uDiffuseAlpha)
    {
        if (s.albedo.a < uDiffuseAlphaTest) discard;
    }

    if (uSpecWorkflow)
    {
        specColor = sRGB2linearV3(specColor); //seems wrong, but it works, weird
        s.reflectivity = lerp(specColor, specColor * resultTint.rgb, saturate(dot(specColor, 1) *  0.5));
        s.albedo.rgb = s.albedo.rgb - s.albedo.rgb * s.reflectivity.rgb;
    }
    else
    {
        vec3 dielectric = vec3(0.04, 0.04, 0.04);
        s.reflectivity = mix(dielectric, s.albedo.rgb, m);
        s.albedo.rgb = s.albedo.rgb - s.albedo.rgb * m;
    }

    //float F0 = maxcomponent(s.reflectivity);
    //s.metalness = remapSpecToMetal(F0);
    //s.eta = remapSpecToEta(F0);

    //hacky roughness filter (rough cavities like in evolution)
    s.gloss = g - (1.0 - nrmMap.b);
    
    //emissive
    if (uUseEmissive)
	{
		vec3 emComp = texture2D(emissiveMap, s.vertexTexCoord).rrr;
		emComp *= emComp;
	
		float emMap = emComp.r;
	
		vec2 emCoords = s.vertexTexCoord + fract(uEmissivePan * fTime(uRT) / 4);
	
	
		float emAtten = uEmissiveAtten;
		if (uEmMask) emAtten *= texture2D(emissiveMap, emCoords.xy).g;
		if (uEmFresnel) emAtten *= pow(saturate(dot(s.normal,s.vertexEye)), uEmFresnelPow);
	
		s.emissiveLight = emComp * emAtten;
		s.emissiveLight *= mix(uEmissiveLow, uEmissiveHigh, emMap);
	}	    
}

//replace Surface subroutine with the custom one
#ifdef Surface
	#undef Surface
#endif
#define Surface WarframePBR

//unused subroutines, all the work is done in the main function
#ifdef Albedo
	#undef Albedo
#endif

#ifdef Microsurface
	#undef Microsurface
#endif

#ifdef Reflectivity
	#undef Reflectivity
#endif

#ifdef Emissive
	#undef Emissive
#endif

#ifdef Transparency
    #undef Transparency
#endif